from typing import Dict, Optional

from jj.mock import HistoryRequest
from multidict import CIMultiDict, CIMultiDictProxy, MultiDict, MultiDictProxy

__all__ = ("make_history_request",)


def make_history_request(*,
                         method: str = "GET",
                         path: str = "/",
                         segments: Optional[Dict[str, str]] = None,
                         params: Optional[MultiDict] = None,
                         headers: Optional[CIMultiDict] = None,
                         body: bytes = b"") -> HistoryRequest:
    if segments is None:
        segments = {}
    if params is None:
        params = MultiDict()
    if headers is None:
        headers = CIMultiDict()

    return HistoryRequest(
        method=method,
        path=path,
        segments=segments,
        params=MultiDictProxy(params),
        headers=CIMultiDictProxy(headers),
        body=body,
    )



from baby_steps import given, then, when
from blahblah import fake

from jj_district42.types.istr import IStrSchema


def test_istr_generation():
    with given:
        sch = IStrSchema()

    with when:
        res = fake(sch)

    with then:
        assert isinstance(res, str)


def test_istr_value_generation():
    with given:
        value = "banana"
        sch = IStrSchema()(value)

    with when:
        res = fake(sch)

    with then:
        assert res in (value, value.lower(), value.upper())



from typing import Any, cast

from district42 import SchemaVisitor
from district42 import SchemaVisitorReturnType as ReturnType
from district42 import schema as sch
from district42.types import DictSchema, GenericTypeAliasSchema, TypeAliasProps
from district42_exp_types.unordered import unordered_schema
from niltype import Nil, Nilable
from revolt import Substitutor
from th import PathHolder
from valera import ValidationResult, Validator

__all__ = ("ParamListSchema", "ParamListProps",
           "ParamListSubstitutor", "ParamListValidator",)


class ParamListProps(TypeAliasProps):
    @property
    def type(self) -> DictSchema:
        key_val = sch.list([sch.str, sch.str])
        type_ = self.get("type", unordered_schema(key_val))
        return cast(DictSchema, type_)


class ParamListSchema(GenericTypeAliasSchema[ParamListProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        try:
            return cast(ReturnType, visitor.visit_jj_param_list(self, **kwargs))
        except AttributeError:
            return visitor.visit_type_alias(self, **kwargs)


class ParamListSubstitutor(Substitutor, extend=True):
    def visit_jj_param_list(self, schema: ParamListSchema, *,
                            value: Any = Nil, **kwargs: Any) -> ParamListSchema:
        if isinstance(value, dict):
            value = [[key, val] for key, val in value.items()]
        return cast(ParamListSchema, self.visit_type_alias(schema, value=value, **kwargs))


class ParamListValidator(Validator, extend=True):
    def visit_jj_param_list(self, schema: ParamListSchema, *,
                            value: Any = Nil, path: Nilable[PathHolder] = Nil,
                            **kwargs: Any) -> ValidationResult:
        if isinstance(value, dict):
            value = [[key, val] for key, val in value.items()]
        return self.visit_type_alias(schema, value=value, path=path, **kwargs)



from baby_steps import then, when

from jj_district42 import HistoryRequestSchema


def test_history_request_declaration():
    with when:
        sch = HistoryRequestSchema()

    with then:
        assert isinstance(sch, HistoryRequestSchema)



from typing import Any

from blahblah import Generator
from jj.mock import HistoryResponse

from ._history_response_schema import HistoryResponseSchema

__all__ = ("HistoryResponseGenerator",)


class HistoryResponseGenerator(Generator, extend=True):
    def visit_jj_history_response(self,
                                  schema: HistoryResponseSchema,
                                  **kwargs: Any) -> HistoryResponse:
        generated = self.visit_type_alias(schema, **kwargs)
        return HistoryResponse(**generated)



from baby_steps import given, then, when
from th import PathHolder
from valera import validate
from valera.errors import MissingElementValidationError, TypeValidationError

from jj_district42.types.param_list import ParamListSchema


def test_param_list_empty_list_validation():
    with given:
        value = []

    with when:
        result = validate(ParamListSchema(), value)

    with then:
        assert result.get_errors() == []


def test_param_list_list_validation():
    with given:
        value = [["key", "val"]]

    with when:
        result = validate(ParamListSchema(), value)

    with then:
        assert result.get_errors() == []


def test_param_list_empty_dict_validation():
    with given:
        value = {}

    with when:
        result = validate(ParamListSchema(), value)

    with then:
        assert result.get_errors() == []


def test_param_list_dict_validation():
    with given:
        value = {"key": "val"}

    with when:
        result = validate(ParamListSchema(), value)

    with then:
        assert result.get_errors() == []


def test_param_list_type_validation_error():
    with given:
        value = set()

    with when:
        result = validate(ParamListSchema(), value)

    with then:
        assert result.get_errors() == [TypeValidationError(PathHolder(), value, list)]


def test_param_list_value_validation_error():
    with given:
        value = [set()]

    with when:
        result = validate(ParamListSchema(), value)

    with then:
        assert result.get_errors() == [TypeValidationError(PathHolder()[0], value[0], list)]


def test_param_list_value_empty_validation_error():
    with given:
        value = [[]]

    with when:
        result = validate(ParamListSchema(), value)

    with then:
        assert result.get_errors() == [MissingElementValidationError(PathHolder()[0], value[0], 0)]


def test_param_list_value_partial_validation_error():
    with given:
        value = [["key"]]

    with when:
        result = validate(ParamListSchema(), value)

    with then:
        assert result.get_errors() == [MissingElementValidationError(PathHolder()[0], value[0], 1)]


def test_param_list_value_key_validation_error():
    with given:
        value = [[None, "val"]]

    with when:
        result = validate(ParamListSchema(), value)

    with then:
        assert result.get_errors() == [TypeValidationError(PathHolder()[0][0], value[0][0], str)]


def test_param_list_value_val_validation_error():
    with given:
        value = [["key", None]]

    with when:
        result = validate(ParamListSchema(), value)

    with then:
        assert result.get_errors() == [TypeValidationError(PathHolder()[0][1], value[0][1], str)]



from baby_steps import given, then, when
from district42 import represent

from jj_district42.types.header_list import HeaderListSchema


def test_header_list_representation():
    with given:
        sch = HeaderListSchema()

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "HeaderListSchema<schema.unordered(schema.list([",
            "    schema.istr,",
            "    schema.str",
            "]))>",
        ])



from ._history_request_generator import HistoryRequestGenerator
from ._history_request_schema import HistoryRequestProps, HistoryRequestSchema, RequestSchema
from ._history_request_substitutor import (
    HistoryRequestSubstitutor,
    HistoryRequestSubstitutorValidator,
)
from ._history_request_validator import HistoryRequestValidator

__all__ = ("HistoryRequestSchema", "HistoryRequestProps", "RequestSchema",
           "HistoryRequestValidator", "HistoryRequestGenerator",
           "HistoryRequestSubstitutor", "HistoryRequestSubstitutorValidator",)



from baby_steps import given, then, when
from district42 import represent

from jj_district42.types.param_list import ParamListSchema


def test_param_list_representation():
    with given:
        sch = ParamListSchema()

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "ParamListSchema<schema.unordered(schema.list([",
            "    schema.str,",
            "    schema.str",
            "]))>",
        ])



from baby_steps import given, then, when
from district42 import schema
from jj.mock import HistoryRequest
from revolt import substitute
from th import PathHolder
from valera import validate
from valera.errors import SchemaMismatchValidationError, TypeValidationError, ValueValidationError

from jj_district42 import HistoryRequestSchema

from ._utils import make_history_request


def test_request_history_type_validation():
    with given:
        sch = HistoryRequestSchema()
        req = make_history_request()

    with when:
        result = validate(sch, req)

    with then:
        assert result.get_errors() == []


def test_request_history_type_validation_error():
    with given:
        sch = HistoryRequestSchema()
        req = make_history_request()
        val = req.to_dict()

    with when:
        result = validate(sch, val)

    with then:
        assert result.get_errors() == [
            TypeValidationError(PathHolder(), val, HistoryRequest)
        ]


def test_request_history_method_validation():
    with given:
        method = "GET"
        sch = substitute(HistoryRequestSchema(), {"method": method})
        req = make_history_request(method=method)

    with when:
        result = validate(sch, req)

    with then:
        assert result.get_errors() == []


def test_request_history_method_validation_error():
    with given:
        expected_method, actual_method = "GET", "POST"
        sch = substitute(HistoryRequestSchema(), {"method": expected_method})
        req = make_history_request(method=actual_method)

    with when:
        result = validate(sch, req)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()["method"], actual_method, expected_method)
        ]


def test_request_history_path_validation():
    with given:
        path = "/users"
        sch = substitute(HistoryRequestSchema(), {"path": path})
        req = make_history_request(path=path)

    with when:
        result = validate(sch, req)

    with then:
        assert result.get_errors() == []


def test_request_history_path_validation_error():
    with given:
        expected_path, actual_path = "/users", "/"
        sch = substitute(HistoryRequestSchema(), {"path": expected_path})
        req = make_history_request(path=actual_path)

    with when:
        result = validate(sch, req)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()["path"], actual_path, expected_path)
        ]


def test_request_history_segments_validation():
    with given:
        segments = {"user_id": "1"}
        sch = substitute(HistoryRequestSchema(), {"segments": segments})
        req = make_history_request(segments=segments)

    with when:
        result = validate(sch, req)

    with then:
        assert result.get_errors() == []


def test_request_history_segments_validation_error():
    with given:
        expected_segments = {"user_id": "1"}
        actual_segments = {"user_id": "2"}
        sch = substitute(HistoryRequestSchema(), {"segments": expected_segments})
        req = make_history_request(segments=actual_segments)

    with when:
        result = validate(sch, req)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()["segments"]["user_id"],
                                 actual_segments["user_id"],
                                 expected_segments["user_id"])
        ]


def test_request_history_body_validation():
    with given:
        body = b""
        sch = substitute(HistoryRequestSchema(), {"body": body})
        req = make_history_request(body=body)

    with when:
        result = validate(sch, req)

    with then:
        assert result.get_errors() == []


def test_request_history_body_validation_error():
    with given:
        expected_body, actual_body = b"<expected>", b"<actual>"
        sch = substitute(HistoryRequestSchema(), {"body": expected_body})
        req = make_history_request(body=actual_body)

    with when:
        result = validate(sch, req)

    with then:
        assert result.get_errors() == [
            SchemaMismatchValidationError(PathHolder()["body"],
                                          actual_body,
                                          (schema.bytes(expected_body),))
        ]


def test_request_history_validation():
    with given:
        req = make_history_request(method="GET", path="/")
        sch = HistoryRequestSchema()

    with when:
        result = validate(sch, req)

    with then:
        assert result.get_errors() == []


def test_request_history_validation_error():
    with given:
        method, path = "GET", "/users"
        sch = substitute(HistoryRequestSchema(), {"method": method, "path": path})
        req = make_history_request(method=method, path="/")

    with when:
        result = validate(sch, req)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()["path"], "/", path)
        ]


def test_request_history_nested():
    with given:
        sch = schema.list([HistoryRequestSchema()])
        req = make_history_request()

    with when:
        result = validate(sch, [req])

    with then:
        assert result.get_errors() == []



from typing import Any

from jj.mock import HistoryRequest
from niltype import Nil, Nilable
from th import PathHolder
from valera import ValidationResult, Validator

from ._history_request_schema import HistoryRequestSchema

__all__ = ("HistoryRequestValidator",)


class HistoryRequestValidator(Validator, extend=True):
    def visit_jj_history_request(self, schema: HistoryRequestSchema, *,
                                 value: Any = Nil, path: Nilable[PathHolder] = Nil,
                                 **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, HistoryRequest):
            return result.add_error(error)

        return self.visit_type_alias(schema, value=value.to_dict(), path=path, **kwargs)



from baby_steps import given, then, when
from blahblah import generate
from jj.mock import HistoryResponse

from jj_district42 import HistoryResponseSchema


def test_history_response_generation():
    with given:
        sch = HistoryResponseSchema()

    with when:
        res = generate(sch)

    with then:
        assert isinstance(res, HistoryResponse)



from typing import Optional

from jj.mock import HistoryResponse
from multidict import CIMultiDict, CIMultiDictProxy

__all__ = ("make_history_response",)


def make_history_response(*,
                          status: int = 200,
                          reason: str = "OK",
                          headers: Optional[CIMultiDict] = None,
                          body: bytes = b"") -> HistoryResponse:
    if headers is None:
        headers = CIMultiDict()
    return HistoryResponse(
        status=status,
        reason=reason,
        headers=CIMultiDictProxy(headers),
        body=body,
    )



from baby_steps import given, then, when
from blahblah import generate
from jj.mock import HistoryRequest

from jj_district42 import HistoryRequestSchema


def test_history_request_generation():
    with given:
        sch = HistoryRequestSchema()

    with when:
        res = generate(sch)

    with then:
        assert isinstance(res, HistoryRequest)



from baby_steps import then, when

from jj_district42 import HistoryResponseSchema


def test_history_response_declaration():
    with when:
        sch = HistoryResponseSchema()

    with then:
        assert isinstance(sch, HistoryResponseSchema)



from baby_steps import given, then, when
from district42 import represent

from jj_district42 import HistoryResponseSchema


def test_history_response_representation():
    with given:
        sch = HistoryResponseSchema()

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "HistoryResponseSchema<schema.dict({",
            "    'status': schema.int,",
            "    'reason': schema.str,",
            "    'headers': HeaderListSchema<schema.unordered(schema.list([",
            "        schema.istr,",
            "        schema.str",
            "    ]))>,",
            "    'body': schema.any",
            "})>",
        ])



from baby_steps import given, then, when
from district42 import schema
from jj.mock import HistoryResponse
from revolt import substitute
from th import PathHolder
from valera import validate
from valera.errors import SchemaMismatchValidationError, TypeValidationError, ValueValidationError

from jj_district42 import HistoryResponseSchema

from ._utils import make_history_response


def test_response_history_type_validation():
    with given:
        sch = HistoryResponseSchema()
        resp = make_history_response()

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == []


def test_response_history_type_validation_error():
    with given:
        sch = HistoryResponseSchema()
        resp = make_history_response()
        val = resp.to_dict()

    with when:
        result = validate(sch, val)

    with then:
        assert result.get_errors() == [
            TypeValidationError(PathHolder(), val, HistoryResponse)
        ]


def test_response_history_status_validation():
    with given:
        status = 200
        sch = substitute(HistoryResponseSchema(), {"status": status})
        resp = make_history_response(status=status)

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == []


def test_response_history_status_validation_error():
    with given:
        expected_status, actual_status = 200, 404
        sch = substitute(HistoryResponseSchema(), {"status": expected_status})
        resp = make_history_response(status=actual_status)

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()["status"], actual_status, expected_status)
        ]


def test_response_history_reason_validation():
    with given:
        reason = "OK"
        sch = substitute(HistoryResponseSchema(), {"reason": reason})
        resp = make_history_response(reason=reason)

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == []


def test_response_history_reason_validation_error():
    with given:
        expected_reason, actual_reason = "OK", "NOT FOUND"
        sch = substitute(HistoryResponseSchema(), {"reason": expected_reason})
        resp = make_history_response(reason=actual_reason)

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()["reason"], actual_reason, expected_reason)
        ]


def test_response_history_body_validation():
    with given:
        body = b"200 OK"
        sch = substitute(HistoryResponseSchema(), {"body": body})
        resp = make_history_response(body=body)

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == []


def test_response_history_body_validation_error():
    with given:
        expected_body, actual_body = b"200 OK", b"404 NOT FOUND"
        sch = substitute(HistoryResponseSchema(), {"body": expected_body})
        resp = make_history_response(body=actual_body)

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == [
            SchemaMismatchValidationError(PathHolder()["body"],
                                          actual_body,
                                          (schema.bytes(expected_body),))
        ]


def test_response_history_validation():
    with given:
        resp = make_history_response(status=200, reason="OK")
        sch = HistoryResponseSchema()

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == []


def test_response_history_validation_error():
    with given:
        status, reason = 200, "OK"
        sch = substitute(HistoryResponseSchema(), {"status": status, "reason": reason})
        resp = make_history_response(status=status, reason="")

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()["reason"], "", reason)
        ]


def test_response_history_nested():
    with given:
        sch = schema.list([HistoryResponseSchema()])
        resp = make_history_response()

    with when:
        result = validate(sch, [resp])

    with then:
        assert result.get_errors() == []



from baby_steps import given, then, when
from district42 import represent

from jj_district42.types.istr import IStrSchema


def test_istr_representation():
    with given:
        sch = IStrSchema()

    with when:
        res = represent(sch)

    with then:
        assert res == "schema.istr"


def test_istr_value_representation():
    with given:
        value = "Banana"
        sch = IStrSchema()(value)

    with when:
        res = represent(sch)

    with then:
        assert res == f"schema.istr({value!r})"



from typing import Callable

import pytest
from baby_steps import given, then, when
from pytest import raises
from revolt import substitute
from revolt.errors import SubstitutionError

from jj_district42.types.istr import IStrSchema


def test_istr_substitution():
    with given:
        value = "banana"
        sch = IStrSchema()

    with when:
        res = substitute(sch, value)

    with then:
        assert res == IStrSchema()(value)
        assert res != sch


@pytest.mark.parametrize("modifier", [str, str.lower, str.upper])
def test_istr_value_substitution(modifier: Callable[[str], str]):
    with given:
        value = "Banana"
        sch = IStrSchema()(value)

    with when:
        res = substitute(sch, modifier(value))

    with then:
        assert res == IStrSchema()(modifier(value))
        assert id(res) != id(sch)


def test_istr_substitution_invalid_value_error():
    with given:
        value = "banana"
        sch = IStrSchema()(value)

    with when, raises(Exception) as exception:
        substitute(sch, ["b", "a", "n", "a", "n", "a"])

    with then:
        assert exception.type is SubstitutionError


def test_istr_substitution_incorrect_value_error():
    with given:
        value = "banana"
        sch = IStrSchema()(value)

    with when, raises(Exception) as exception:
        substitute(sch, "apple")

    with then:
        assert exception.type is SubstitutionError



import random
from typing import Any, cast

from blahblah import Generator
from district42 import Props, SchemaVisitor
from district42 import SchemaVisitorReturnType as ReturnType
from district42.errors import make_already_declared_error, make_invalid_type_error
from district42.representor import Representor
from district42.types import Schema
from niltype import Nil, Nilable
from revolt import Substitutor
from revolt.errors import make_substitution_error
from th import PathHolder
from valera import ValidationResult, Validator
from valera.errors import ValueValidationError

__all__ = ("IStrSchema", "IStrProps",
           "IStrRepresentor", "IStrGenerator", "IStrValidator", "IStrSubstitutor",)


class IStrProps(Props):
    @property
    def value(self) -> Nilable[str]:
        return self.get("value")


class IStrSchema(Schema[IStrProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        return cast(ReturnType, visitor.visit_istr(self, **kwargs))

    def __call__(self, /, value: str) -> "IStrSchema":
        if not isinstance(value, str):
            raise make_invalid_type_error(self, value, (str,))

        if self.props.value is not Nil:
            raise make_already_declared_error(self)

        return self.__class__(self.props.update(value=value))


class IStrRepresentor(Representor, extend=True):
    def visit_istr(self, schema: IStrSchema, *, indent: int = 0, **kwargs: Any) -> str:
        r = f"{self._name}.istr"
        if schema.props.value is not Nil:
            r += f"({schema.props.value!r})"
        return r


class IStrGenerator(Generator, extend=True):
    def visit_istr(self, schema: IStrSchema, **kwargs: Any) -> str:
        if schema.props.value is not Nil:
            modifier = random.choice((str, str.lower, str.upper))
            return modifier(schema.props.value)  # type: ignore
        return ""


class IStrValidator(Validator, extend=True):
    def visit_istr(self, schema: IStrSchema, *,
                   value: Any = Nil, path: Nilable[PathHolder] = Nil,
                   **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, str):
            return result.add_error(error)

        if schema.props.value is not Nil:
            if value.lower() != schema.props.value.lower():
                error = ValueValidationError(path, value, schema.props.value)
                return result.add_error(error)

        return result


class IStrSubstitutor(Substitutor, extend=True):
    def visit_istr(self, schema: IStrSchema, *,
                   value: Any = Nil, **kwargs: Any) -> IStrSchema:
        result = schema.__accept__(self._validator, value=value)
        if result.has_errors():
            raise make_substitution_error(result, self._formatter)
        return schema.__class__(schema.props.update(value=value))



from _pytest.python_api import raises
from baby_steps import given, then, when
from district42 import schema
from multidict import CIMultiDict
from revolt import substitute
from revolt.errors import SubstitutionError

from jj_district42 import HistoryResponseSchema
from jj_district42.history_response import ResponseSchema

from ._utils import make_history_response


def test_history_response_empty_dict_substitution():
    with given:
        sch = HistoryResponseSchema()

    with when:
        res = substitute(sch, {})

    with then:
        assert isinstance(res, HistoryResponseSchema)
        assert id(res) != id(sch)


def test_history_response_incorrect_type_substitution_error():
    with given:
        sch = HistoryResponseSchema()

    with when, raises(Exception) as exception:
        substitute(sch, set())

    with then:
        assert exception.type is SubstitutionError


def test_history_response_dict_substitution():
    with given:
        resp = {
            "status": 200,
            "reason": "OK",
            "headers": {"Authorization": "banana"},
            "body": b"[]",
        }
        sch = HistoryResponseSchema()

    with when:
        res = substitute(sch, resp)

    with then:
        assert res.props.type == substitute(ResponseSchema, resp)
        assert id(res) != id(sch)


def test_history_response_resp_substitution():
    with given:
        resp = {
            "status": 200,
            "reason": "OK",
            "headers": {"Authorization": "banana"},
            "body": b"[]",
        }
        history_response = make_history_response(
            status=resp["status"],
            reason=resp["reason"],
            headers=CIMultiDict(resp["headers"]),
            body=resp["body"],
        )
        sch = HistoryResponseSchema()

    with when:
        res = substitute(sch, history_response)

    with then:
        assert res.props.type == substitute(ResponseSchema, resp)
        assert id(res) != id(sch)


def test_history_response_inner_dict_substitution():
    with given:
        sch = schema.dict({
            "response": HistoryResponseSchema()
        })

    with when:
        res = substitute(sch, {
            "response": {}
        })

    with then:
        assert isinstance(res["response"], HistoryResponseSchema)
        assert id(res) != id(sch)


def test_history_response_inner_resp_substitution():
    with given:
        sch = schema.dict({
            "response": HistoryResponseSchema()
        })
        resp = make_history_response()

    with when:
        res = substitute(sch, {
            "response": resp
        })

    with then:
        assert isinstance(res["response"], HistoryResponseSchema)
        assert id(res) != id(sch)



from typing import Any, cast

from district42 import SchemaVisitor
from district42 import SchemaVisitorReturnType as ReturnType
from district42 import schema
from district42.types import DictSchema, GenericTypeAliasSchema, TypeAliasProps

from jj_district42.types.header_list import HeaderListSchema

__all__ = ("HistoryResponseSchema", "HistoryResponseProps", "ResponseSchema",)


ResponseSchema = schema.dict({
    "status": schema.int,
    "reason": schema.str,
    "headers": HeaderListSchema(),
    "body": schema.any
})


class HistoryResponseProps(TypeAliasProps):
    @property
    def type(self) -> DictSchema:
        return cast(DictSchema, self.get("type", ResponseSchema))


class HistoryResponseSchema(GenericTypeAliasSchema[HistoryResponseProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        try:
            return cast(ReturnType, visitor.visit_jj_history_response(self, **kwargs))
        except AttributeError:
            return visitor.visit_type_alias(self, **kwargs)



from district42 import register_type, schema
from district42_exp_types.uuid_str import UUIDStrSchema

from ._version import version
from .history_request import HistoryRequestSchema
from .history_response import HistoryResponseSchema

schema_history_request = register_type("jj_history_request", HistoryRequestSchema)
schema_history_response = register_type("jj_history_response", HistoryResponseSchema)

HistoryItemSchema = schema.dict({
    "request": HistoryRequestSchema(),
    "response": HistoryResponseSchema(),
    "tags": schema.list(UUIDStrSchema()),
})

HistorySchema = schema.list(HistoryItemSchema)

__version__ = version
__all__ = ("HistorySchema", "HistoryItemSchema",
           "schema_history_request", "schema_history_response",
           "HistoryRequestSchema", "HistoryResponseSchema",)



from uuid import uuid4

from baby_steps import given, then, when
from district42.errors import DeclarationError
from pytest import raises

from jj_district42.types.istr import IStrSchema


def test_istr_declaration():
    with when:
        sch = IStrSchema()

    with then:
        assert isinstance(sch, IStrSchema)


def test_istr_value_declaration():
    with given:
        value = str(uuid4())

    with when:
        sch = IStrSchema()(value)

    with then:
        assert sch.props.value == value


def test_istr_invalid_value_type_declaration_error():
    with given:
        value = ["b", "a", "n", "a", "n", "a"]

    with when, raises(Exception) as exception:
        IStrSchema()(value)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == ("`schema.istr` value must be an instance of 'str', "
                                        f"instance of 'list' {value!r} given")


def test_istr_already_declared_declaration_error():
    with given:
        value = "banana"
        another_value = "apple"

    with when, raises(Exception) as exception:
        IStrSchema()(value)(another_value)

    with then:
        assert exception.type is DeclarationError
        assert str(exception.value) == f"`schema.istr({value!r})` is already declared"



from baby_steps import then, when

from jj_district42.types.header_list import HeaderListSchema


def test_header_list_declaration():
    with when:
        sch = HeaderListSchema()

    with then:
        assert isinstance(sch, HeaderListSchema)



from baby_steps import given, then, when
from blahblah import fake

from jj_district42.types.param_list import ParamListSchema


def test_param_list_generation():
    with given:
        sch = ParamListSchema()

    with when:
        res = fake(sch)

    with then:
        assert isinstance(res, list)



from baby_steps import given, then, when
from district42 import represent

from jj_district42 import HistoryRequestSchema


def test_history_request_representation():
    with given:
        sch = HistoryRequestSchema()

    with when:
        res = represent(sch)

    with then:
        assert res == "\n".join([
            "HistoryRequestSchema<schema.dict({",
            "    'method': schema.str.len(1, ...),",
            "    'path': schema.str,",
            "    'segments': schema.dict,",
            "    'params': ParamListSchema<schema.unordered(schema.list([",
            "        schema.str,",
            "        schema.str",
            "    ]))>,",
            "    'headers': HeaderListSchema<schema.unordered(schema.list([",
            "        schema.istr,",
            "        schema.str",
            "    ]))>,",
            "    'body': schema.any",
            "})>",
        ])



from typing import Any

from blahblah import Generator
from jj.mock import HistoryRequest

from ._history_request_schema import HistoryRequestSchema

__all__ = ("HistoryRequestGenerator",)


class HistoryRequestGenerator(Generator, extend=True):
    def visit_jj_history_request(self,
                                 schema: HistoryRequestSchema,
                                 **kwargs: Any) -> HistoryRequest:
        generated = self.visit_type_alias(schema, **kwargs)
        return HistoryRequest(**generated)



from typing import Any, cast

from district42 import SchemaVisitor
from district42 import SchemaVisitorReturnType as ReturnType
from district42 import schema
from district42.types import DictSchema, GenericTypeAliasSchema, TypeAliasProps

from jj_district42.types.header_list import HeaderListSchema
from jj_district42.types.param_list import ParamListSchema

__all__ = ("HistoryRequestSchema", "HistoryRequestProps", "RequestSchema",)


RequestSchema = schema.dict({
    "method": schema.str.len(1, ...),
    "path": schema.str,
    "segments": schema.dict,
    "params": ParamListSchema(),
    "headers": HeaderListSchema(),
    "body": schema.any,
})


class HistoryRequestProps(TypeAliasProps):
    @property
    def type(self) -> DictSchema:
        return cast(DictSchema, self.get("type", RequestSchema))


class HistoryRequestSchema(GenericTypeAliasSchema[HistoryRequestProps]):
    def __accept__(self, visitor: SchemaVisitor[ReturnType], **kwargs: Any) -> ReturnType:
        try:
            return cast(ReturnType, visitor.visit_jj_history_request(self, **kwargs))
        except AttributeError:
            return visitor.visit_type_alias(self, **kwargs)



from typing import Callable

import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder
from valera import validate
from valera.errors import TypeValidationError, ValueValidationError

from jj_district42.types.istr import IStrSchema


@pytest.mark.parametrize("modifier", [str, str.lower, str.upper])
def test_istr_type_validation(modifier: Callable[[str], str]):
    with given:
        value = modifier("Banana")

    with when:
        result = validate(IStrSchema(), value)

    with then:
        assert result.get_errors() == []


def test_istr_type_validation_error():
    with given:
        value = ["b", "a", "n", "a", "n", "a"]

    with when:
        result = validate(IStrSchema(), value)

    with then:
        assert result.get_errors() == [TypeValidationError(PathHolder(), value, str)]


@pytest.mark.parametrize(("actual_func", "expected_func"), [
    (str.lower, str.upper),
    (str.upper, str.lower),
])
def test_istr_value_validation(actual_func: Callable[[str], str],
                               expected_func: Callable[[str], str]):
    with given:
        value = "banana"
        sch = IStrSchema()(expected_func(value))

    with when:
        result = validate(sch, actual_func(value))

    with then:
        assert result.get_errors() == []


def test_istr_value_validation_error():
    with given:
        expected_value = "banana"
        actual_value = "apple"

    with when:
        result = validate(IStrSchema()(expected_value), actual_value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder(), actual_value, expected_value)
        ]


def test_istr_nested_validation():
    with given:
        value = "banana"
        sch = schema.list([IStrSchema()(value)])

    with when:
        result = validate(sch, [value])

    with then:
        assert result.get_errors() == []



from _pytest.python_api import raises
from baby_steps import given, then, when
from district42 import schema
from multidict import CIMultiDict, MultiDict
from revolt import substitute
from revolt.errors import SubstitutionError

from jj_district42 import HistoryRequestSchema
from jj_district42.history_request import RequestSchema

from ._utils import make_history_request


def test_history_request_empty_dict_substitution():
    with given:
        sch = HistoryRequestSchema()

    with when:
        res = substitute(sch, {})

    with then:
        assert isinstance(res, HistoryRequestSchema)
        assert id(res) != id(sch)


def test_history_request_incorrect_type_substitution_error():
    with given:
        sch = HistoryRequestSchema()

    with when, raises(Exception) as exception:
        substitute(sch, set())

    with then:
        assert exception.type is SubstitutionError


def test_history_request_dict_substitution():
    with given:
        req = {
            "method": "GET",
            "path": "/users",
            "segments": {},
            "params": {"user_id": "1"},
            "headers": {"Authorization": "banana"},
            "body": b"[]",
        }
        sch = HistoryRequestSchema()

    with when:
        res = substitute(sch, req)

    with then:
        assert res.props.type == substitute(RequestSchema, req)
        assert id(res) != id(sch)


def test_history_request_req_substitution():
    with given:
        req = {
            "method": "GET",
            "path": "/users",
            "segments": {},
            "params": {"user_id": "1"},
            "headers": {"Authorization": "banana"},
            "body": b"[]",
        }
        history_request = make_history_request(
            method=req["method"],
            path=req["path"],
            segments=req["segments"],
            params=MultiDict(req["params"]),
            headers=CIMultiDict(req["headers"]),
            body=req["body"],
        )
        sch = HistoryRequestSchema()

    with when:
        res = substitute(sch, history_request)

    with then:
        assert res.props.type == substitute(RequestSchema, req)
        assert id(res) != id(sch)


def test_history_request_inner_dict_substitution():
    with given:
        sch = schema.dict({
            "request": HistoryRequestSchema()
        })

    with when:
        res = substitute(sch, {
            "request": {}
        })

    with then:
        assert isinstance(res["request"], HistoryRequestSchema)
        assert id(res) != id(sch)


def test_history_request_inner_req_substitution():
    with given:
        sch = schema.dict({
            "request": HistoryRequestSchema()
        })
        req = make_history_request()

    with when:
        res = substitute(sch, {
            "request": req
        })

    with then:
        assert isinstance(res["request"], HistoryRequestSchema)
        assert id(res) != id(sch)



from ._history_response_generator import HistoryResponseGenerator
from ._history_response_schema import HistoryResponseProps, HistoryResponseSchema, ResponseSchema
from ._history_response_substitutor import (
    HistoryResponseSubstitutor,
    HistoryResponseSubstitutorValidator,
)
from ._history_response_validator import HistoryResponseValidator

__all__ = ("HistoryResponseSchema", "HistoryResponseProps", "ResponseSchema",
           "HistoryResponseValidator", "HistoryResponseGenerator",
           "HistoryResponseSubstitutor", "HistoryResponseSubstitutorValidator",)



from typing import Any, cast

from jj.mock import HistoryResponse
from niltype import Nil, Nilable
from revolt import Substitutor, SubstitutorValidator
from th import PathHolder
from valera import ValidationResult

from ._history_response_schema import HistoryResponseSchema

__all__ = ("HistoryResponseSubstitutor", "HistoryResponseSubstitutorValidator",)


class HistoryResponseSubstitutor(Substitutor, extend=True):
    def visit_jj_history_response(self, schema: HistoryResponseSchema, *,
                                  value: Any = Nil, **kwargs: Any) -> HistoryResponseSchema:
        if isinstance(value, HistoryResponse):
            value = value.to_dict()
        return cast(HistoryResponseSchema, self.visit_type_alias(schema, value=value, **kwargs))


class HistoryResponseSubstitutorValidator(SubstitutorValidator, extend=True):
    def visit_jj_history_response(self, schema: HistoryResponseSchema, *,
                                  value: Any = Nil, path: Nilable[PathHolder] = Nil,
                                  **kwargs: Any) -> ValidationResult:
        if isinstance(value, HistoryResponse):
            value = value.to_dict()
        return self.visit_type_alias(schema, value=value, path=path, **kwargs)



from baby_steps import given, then, when
from district42 import schema
from district42_exp_types.unordered import UnorderedContainsValidationError
from multidict import CIMultiDict
from revolt import substitute
from th import PathHolder
from valera import validate
from valera.errors import ExtraElementValidationError

from jj_district42 import HistoryResponseSchema
from jj_district42.types.istr import IStrSchema

from ._utils import make_history_response


def test_response_history_no_headers_validation():
    with given:
        sch = HistoryResponseSchema()
        headers = {"user_id": "1"}
        resp = make_history_response(headers=CIMultiDict(headers))

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == []


def test_response_history_dict_headers_validation():
    with given:
        headers = {"user_id": "1"}
        sch = substitute(HistoryResponseSchema(), {"headers": headers})
        resp = make_history_response(headers=CIMultiDict(headers))

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == []


def test_response_history_dict_headers_case_validation():
    with given:
        actual_headers = {"User_Id": "1"}
        expected_headers = {"user_id": "1"}
        sch = substitute(HistoryResponseSchema(), {"headers": expected_headers})
        resp = make_history_response(headers=CIMultiDict(actual_headers))

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == []


def test_response_history_list_headers_validation():
    with given:
        headers = [["user_id", "1"]]
        sch = substitute(HistoryResponseSchema(), {"headers": headers})
        resp = make_history_response(headers=CIMultiDict(headers))

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == []


def test_response_history_list_multiple_headers_validation():
    with given:
        headers = [["user_id", "1"], ["user_id", "2"]]
        sch = substitute(HistoryResponseSchema(), {"headers": headers})
        resp = make_history_response(headers=CIMultiDict(headers))

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == []


def test_response_history_dict_headers_validation_error():
    with given:
        expected_headers = {"user_id": "1"}
        actual_headers = {"user_id": "2"}
        sch = substitute(HistoryResponseSchema(), {"headers": expected_headers})
        resp = make_history_response(headers=CIMultiDict(actual_headers))

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == [
            UnorderedContainsValidationError(PathHolder()["headers"], schema.list([
                IStrSchema()("user_id"),
                schema.str("1")
            ])),
            ExtraElementValidationError(PathHolder()["headers"], ["user_id", "2"], 0)
        ]


def test_response_history_list_headers_validation_error():
    with given:
        expected_headers = [["user_id", "1"]]
        actual_headers = [["user_id", "2"]]
        sch = substitute(HistoryResponseSchema(), {"headers": expected_headers})
        resp = make_history_response(headers=CIMultiDict(actual_headers))

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == [
            UnorderedContainsValidationError(PathHolder()["headers"], schema.list([
                IStrSchema()("user_id"),
                schema.str("1")
            ])),
            ExtraElementValidationError(PathHolder()["headers"], ["user_id", "2"], 0),
        ]


def test_response_history_list_multiple_headers_validation_error():
    with given:
        expected_headers = [["user_id", "1"], ["user_id", "2"]]
        actual_headers = [["user_id", "1"]]
        sch = substitute(HistoryResponseSchema(), {"headers": expected_headers})
        resp = make_history_response(headers=CIMultiDict(actual_headers))

    with when:
        result = validate(sch, resp)

    with then:
        assert result.get_errors() == [
            UnorderedContainsValidationError(PathHolder()["headers"], schema.list([
                IStrSchema()("user_id"),
                schema.str("2")
            ])),
        ]



from _pytest.python_api import raises
from baby_steps import given, then, when
from district42 import schema
from district42_exp_types.unordered import unordered_schema
from revolt import substitute
from revolt.errors import SubstitutionError

from jj_district42.types.param_list import ParamListSchema


def test_param_list_empty_list_substitution():
    with given:
        sch = ParamListSchema()

    with when:
        res = substitute(sch, [])

    with then:
        assert res.props.type == unordered_schema([])
        assert res != sch


def test_param_list_list_substitution():
    with given:
        value = [["key", "val"]]
        sch = ParamListSchema()

    with when:
        res = substitute(sch, value)

    with then:
        assert res.props.type == unordered_schema([
            schema.list([schema.str("key"), schema.str("val")])
        ])
        assert res != sch


def test_param_list_list_multiple_substitution():
    with given:
        value = [["key", "val"], ["key", "another_val"]]
        sch = ParamListSchema()

    with when:
        res = substitute(sch, value)

    with then:
        assert res.props.type == unordered_schema([
            schema.list([schema.str("key"), schema.str("val")]),
            schema.list([schema.str("key"), schema.str("another_val")])
        ])
        assert res != sch


def test_param_list_empty_dict_substitution():
    with given:
        sch = ParamListSchema()

    with when:
        res = substitute(sch, {})

    with then:
        assert res.props.type == unordered_schema([])
        assert res != sch


def test_param_list_dict_substitution():
    with given:
        value = {"key": "val"}
        sch = ParamListSchema()

    with when:
        res = substitute(sch, value)

    with then:
        assert res.props.type == unordered_schema([
            schema.list([schema.str("key"), schema.str("val")])
        ])
        assert res != sch


def test_param_list_dict_multiple_substitution():
    with given:
        value = {"key": "val", "another_key": "val"}
        sch = ParamListSchema()

    with when:
        res = substitute(sch, value)

    with then:
        assert res.props.type == unordered_schema([
            schema.list([schema.str("key"), schema.str("val")]),
            schema.list([schema.str("another_key"), schema.str("val")])
        ])
        assert res != sch


def test_param_list_incorrect_type_substitution_error():
    with given:
        sch = ParamListSchema()

    with when, raises(Exception) as exception:
        substitute(sch, set())

    with then:
        assert exception.type is SubstitutionError


def test_param_list_incorrect_dict_value_substitution_error():
    with given:
        sch = ParamListSchema()

    with when, raises(Exception) as exception:
        substitute(sch, {"key": None})

    with then:
        assert exception.type is SubstitutionError


def test_param_list_incorrect_list_value_substitution_error():
    with given:
        sch = ParamListSchema()

    with when, raises(Exception) as exception:
        substitute(sch, [[]])

    with then:
        assert exception.type is SubstitutionError



from baby_steps import given, then, when
from blahblah import fake

from jj_district42.types.header_list import HeaderListSchema


def test_header_list_generation():
    with given:
        sch = HeaderListSchema()

    with when:
        res = fake(sch)

    with then:
        assert isinstance(res, list)



from baby_steps import then, when

from jj_district42.types.param_list import ParamListSchema


def test_param_list_declaration():
    with when:
        sch = ParamListSchema()

    with then:
        assert isinstance(sch, ParamListSchema)



from typing import Any, cast

from jj.mock import HistoryRequest
from niltype import Nil, Nilable
from revolt import Substitutor, SubstitutorValidator
from th import PathHolder
from valera import ValidationResult

from ._history_request_schema import HistoryRequestSchema

__all__ = ("HistoryRequestSubstitutor", "HistoryRequestSubstitutorValidator",)


class HistoryRequestSubstitutor(Substitutor, extend=True):
    def visit_jj_history_request(self, schema: HistoryRequestSchema, *,
                                 value: Any = Nil, **kwargs: Any) -> HistoryRequestSchema:
        if isinstance(value, HistoryRequest):
            value = value.to_dict()
        return cast(HistoryRequestSchema, self.visit_type_alias(schema, value=value, **kwargs))


class HistoryRequestSubstitutorValidator(SubstitutorValidator, extend=True):
    def visit_jj_history_request(self, schema: HistoryRequestSchema, *,
                                 value: Any = Nil, path: Nilable[PathHolder] = Nil,
                                 **kwargs: Any) -> ValidationResult:
        if isinstance(value, HistoryRequest):
            value = value.to_dict()
        return self.visit_type_alias(schema, value=value, path=path, **kwargs)
